/*	SoundRecorder.h
 * copyright 2002, OpenBeOS project
 * Francois Revol, revol@free.fr
 * derived from HardDiskRec from Marcus Overhagen.
 * released under MIT license
 */

#if !defined(_SOUND_RECORDER_H)
#define _SOUND_RECORDER_H


#include <MediaDefs.h>
#include <MediaNode.h>
//#include <BufferProducer.h>
#include <Locker.h>

//class BMediaTrack;

// use with:
// using namespace BPrivate::Media::Experimental;

namespace BPrivate {
namespace Media {
namespace Experimental {

	// The media node that goes with the object.
	class _SoundRecorderNode;
	// this class should stay in BPrivate::Media

class BSoundRecorder {
public:
		//	This is a common hook function interface for
		//	SoundConsumer and SoundProducer to use.
		typedef void (*SoundProcessFunc)(void * cookie,
			bigtime_t timestamp, void * data, size_t datasize,
			const media_multi_audio_format & format);
		typedef void (*SoundNotifyFunc)(void * cookie,
			int32 code, ...);
		// XXX: move this
		
		//	These are special codes that we use in the Notify
		//	function hook.
		enum sound_recorder_notification {
			B_WILL_START = 1,		//	performance_time
			B_WILL_STOP,			//	performance_time immediate
			B_WILL_SEEK,			//	performance_time media_time
			B_WILL_TIMEWARP,		//	real_time performance_time
			B_CONNECTED,			//	name (char*)
			B_DISCONNECTED,			//
			B_FORMAT_CHANGED,		//	media_multi_audio_format*
			B_NODE_DIES,			//	node will die!
			B_HOOKS_CHANGED,		//	
			B_OP_TIMED_OUT,			//	timeout that expired -- Consumer only
			B_PRODUCER_DATA_STATUS,	//	status performance_time -- Consumer only
			B_LATE_NOTICE			//	how_much performance_time -- Producer only
		};

/*		BSoundRecorder(
				const char * name = NULL,
				BMediaTrack *track,
				SoundProcessFunc recordFunc = NULL,
				SoundNotifyFunc notifyFunc = NULL,
				void * cookie = NULL);
*/

		virtual	~BSoundRecorder();
public:
		BSoundRecorder(
				media_multi_audio_format *format, 
				bool waitForInputConnection = false,
				const char *name = NULL,
				SoundProcessFunc recordFunc = NULL,
				SoundNotifyFunc notifyFunc = NULL,
				void *cookie = NULL);
/*		BSoundRecorder(
				const media_node & fromNode,
				const char * name = NULL,
				BMediaTrack *track,
				const media_output * output = NULL,
				SoundProcessFunc recordFunc = NULL,
				SoundNotifyFunc notifyFunc = NULL,
				void * cookie = NULL);
*/
		BSoundRecorder(
				const media_node &fromNode,
				media_output *output = NULL,
				media_multi_audio_format *format = NULL, 
				const char *name = NULL,
				SoundProcessFunc recordFunc = NULL,
				SoundNotifyFunc notifyFunc = NULL,
				void *cookie = NULL);

		status_t InitCheck();
		media_multi_audio_format Format() const{ return m_input.format.u.raw_audio; };	//	new in R4.1

		status_t Start();
		void Stop(
				bool block = true, // UNIMPLEMENTED
				bool flush = true); //

		SoundProcessFunc BufferRecorder() const;
		void SetBufferRecorder(SoundProcessFunc recordFunc);

		SoundNotifyFunc EventNotifier() const;
		void SetNotifier(SoundNotifyFunc notifyFunc);

		void * Cookie() const;
		void SetCookie(
			void * cookie);
		void SetCallbacks(		//	atomic change
				SoundProcessFunc recordFunc = NULL,
				SoundNotifyFunc notifyFunc = NULL,
				void * cookie = NULL);

		virtual bigtime_t	CurrentTime();
		virtual bigtime_t	PerformanceTime();
		virtual status_t	Preroll();
		virtual bigtime_t	Latency();

		float 		Volume();			/* 0 - 1.0 */
		void 		SetVolume(			/* 0 - 1.0 */
						float new_volume);
		float		VolumeDB(			/* -xx - +xx (see GetVolumeInfo()) */
						bool forcePoll = false);	/* if false, cached value will be used if new enough */
		void		SetVolumeDB(
						float volume_dB);
		status_t	GetVolumeInfo(
						media_node * out_node,
						int32 * out_parameter,
						float * out_min_dB,
						float * out_max_dB);

virtual	bool HasData();
		void SetHasData(		//	this does more than just set a flag
				bool has_data);

		// if waitForInputConnection = true in constructor
		status_t WaitForIncomingConnection(
				media_multi_audio_format *format = NULL,
				bool block = true
				/*const media_output *output = NULL*/);


protected:

		void SetInitError(status_t in_error);

virtual void Notify(
				sound_recorder_notification what,
				...);
virtual void RecordBuffer(
				bigtime_t time,
				void * buffer,
				size_t size,
				const media_multi_audio_format & format);

private:
friend class BPrivate::Media::Experimental::_SoundRecorderNode;

virtual	status_t _Reserved_SoundRecorder_0(void *, ...);
virtual	status_t _Reserved_SoundRecorder_1(void *, ...);
virtual	status_t _Reserved_SoundRecorder_2(void *, ...);
virtual	status_t _Reserved_SoundRecorder_3(void *, ...);
virtual	status_t _Reserved_SoundRecorder_4(void *, ...);
virtual	status_t _Reserved_SoundRecorder_5(void *, ...);
virtual	status_t _Reserved_SoundRecorder_6(void *, ...);
virtual	status_t _Reserved_SoundRecorder_7(void *, ...);
 
		BPrivate::Media::Experimental::_SoundRecorderNode * _m_node;
		SoundProcessFunc _RecordFunc;
		SoundNotifyFunc _NotifyFunc;
		void * _m_cookie;

		BLocker _m_lock;
		float _m_volume;
		media_input m_input;
		media_output m_output;
		media_node m_connectedTo;
//		void * _m_buf;
//		size_t _m_bufsize;
//		int32 _m_has_data;
		sem_id _m_gotConnection;

		bool m_waitingForConnection;
		bool m_isConnected;
		bool m_isRecording;

		status_t _m_init_err;
		bigtime_t _m_perfTime;
		BContinuousParameter * _m_volumeSlider;
		bigtime_t _m_gotVolume;
		BMediaRoster *m_roster;
		char m_name[B_OS_NAME_LENGTH];
		uint32 _m_reserved[10];

		void get_volume_slider();
		status_t Init(
				const media_node *node,
				media_output *output,
				media_multi_audio_format *format,
				bool waitForInputConnection, 
				const char *name,
				SoundProcessFunc recordFunc,
				SoundNotifyFunc notifyFunc,
				void *cookie);
		status_t BSoundRecorder::MakeRecordConnection(
				const media_node & input);
		status_t BSoundRecorder::BreakRecordConnection();
};

} // Experimental
} // Media
} // BPrivate

#endif /* _SOUND_RECORDER_H */
