/**
 * \file MediaUtils.h
 * \brief Media related functions.
 * \author Marcus Overhagen, Francois Revol
 * Copyright 2003, OpenBeOS.
 */

#ifndef __MediaUtil_h
#define __MediaUtil_h

#include <DataIO.h>
#include <MediaDefs.h>
#include <RealtimeAlloc.h>

// use with:
// using namespace BPrivate::Media::Experimental;

namespace BPrivate {
namespace Media {
namespace Experimental {

class BAudioRingBuffer : public BDataIO {
public:
	BAudioRingBuffer(size_t size, bool rtAlloc=true, rtm_pool *fromRTPool=NULL);
	~BAudioRingBuffer();
	status_t	InitCheck() const;

	virtual ssize_t	Read(void *buffer, size_t size);
	virtual ssize_t	Write(const void *buffer, size_t size);

	size_t Queued() const; /* number of bytes queued here */
private:
	size_t fDataLen;
	void *fData;
	rtm_pool *fPool;
	status_t initStatus;
	
	sem_id input_sem;
    int input_index;
    sem_id output_sem;
    int output_index;
};

}; // namespace Experimental
}; // namespace Media
}; // namespace BPrivate

#endif /* __MediaUtil_h */
