#include <Debug.h>
#include <NodeInfo.h>
#include <MediaFile.h>
#include <MediaTrack.h>
#include <MediaNode.h>
#include <MediaRoster.h>
#include <TimeSource.h>
#include <String.h>
#include <Entry.h>
#include <stdio.h>

#include <SoundRecorder.h>

#include "SoundRecorderNode.h"

using namespace BPrivate::Media::Experimental;

/* fool BeIDE function parser */
#define IN_NAMESPACE \
namespace BPrivate { \
namespace Media { \
namespace Experimental {
IN_NAMESPACE
#undef IN_NAMESPACE

#define BSNDREC "BSoundRecorder:"

BSoundRecorder::BSoundRecorder(
				media_multi_audio_format *format, 
				bool waitForInputConnection,
				const char *name,
				SoundProcessFunc recordFunc,
				SoundNotifyFunc notifyFunc,
				void * cookie)
{
	Init(NULL, NULL, format, waitForInputConnection, name, recordFunc, notifyFunc, cookie);
}

BSoundRecorder::BSoundRecorder(
				const media_node &fromNode,
				media_output *output,
				media_multi_audio_format *format, 
				const char *name,
				SoundProcessFunc recordFunc,
				SoundNotifyFunc notifyFunc,
				void * cookie)
{
	Init(&fromNode, output, format, false, name, recordFunc, notifyFunc, cookie);
}

status_t BSoundRecorder::Init(
				const media_node *fromNode,
				media_output *output,
				media_multi_audio_format *format, 
				bool waitForInputConnection,
				const char *name,
				SoundProcessFunc recordFunc,
				SoundNotifyFunc notifyFunc,
				void *cookie)
{
	// init class member
	_m_node = NULL;
	_RecordFunc = recordFunc;
	_NotifyFunc = notifyFunc;
	_m_cookie = cookie;
	_m_volume = 0.0;
//	m_input;
//	m_output;
//	_m_buf = NULL;
//	_m_bufsize = 0;
//	_m_has_data = 0;
	_m_gotConnection = -1;

	_m_init_err = EPIPE;		//	new in R4.1
	_m_perfTime = 0LL;
	_m_volumeSlider = NULL;
	_m_gotVolume = 0LL;

	m_waitingForConnection = waitForInputConnection;
	m_isConnected = false;
	m_isRecording = false;

	status_t error;
	
	if (fromNode)
		m_connectedTo = *fromNode;
	
	if (!name)
		name = "a BSoundRecorder";
	strncpy(m_name, name, B_OS_NAME_LENGTH);
	
	
	
	//	Make sure the media roster is there (which means the server is there).
	BMediaRoster *m_roster = BMediaRoster::Roster(&error);
	if (!m_roster) {
		PRINT((BSNDREC"::Init(): Can't get MediaRoster !\n"));
		goto bad_mojo;
	}

	if (!m_waitingForConnection && !fromNode) {
		error = m_roster->GetAudioInput(&m_connectedTo);
		if (error < B_OK) {	//	there's no input?
			PRINT((BSNDREC"::Init(): Can't get default audio input !\n"));
			goto bad_mojo;
		}
	}

/*	error = m_roster->GetAudioMixer(&m_audioMixerNode);
	if (error < B_OK) {	//	there's no mixer?
		goto bad_mojo;
	}
*/
	//	Create our internal Node which records sound, and register it.
	_m_node = new _SoundRecorderNode(m_name,/* m_format,*/ this);
	error = m_roster->RegisterNode(_m_node);
	if (error < B_OK) {
		PRINT((BSNDREC"::Init(): Can't register recorder node !\n"));
		goto bad_mojo;
	}
	if (!m_waitingForConnection) {
		MakeRecordConnection(m_connectedTo);
		if (format)
			*format = m_input.format.u.raw_audio;
	}
	SetInitError(B_OK);
	
	return B_OK;

	//	Error handling.
bad_mojo:
	if (_m_node) {
		// Release() returns NULL if refcount is 0 (in which case the node it deleted)
		if (_m_node->Release() == NULL)
			_m_node = NULL;
	}
	return B_ERROR;
}

BSoundRecorder::~BSoundRecorder()
{
	if (_m_node) {
		// Release() returns NULL if refcount is 0 (in which case the node it deleted)
		// Here we want to make sure it is deleted.
		// XXX: sure of that ?
		while(_m_node->Release());
	}
}

status_t BSoundRecorder::InitCheck()
{
	return _m_init_err;
}

status_t BSoundRecorder::WaitForIncomingConnection(
		media_multi_audio_format *format,
		bool block
		/*const media_output *output*/)
{
	sem_id sem;
	status_t err;
	PRINT((BSNDREC"::WaitForIncomingConnection()\n"));
	if (!m_waitingForConnection) {
		PRINT((BSNDREC"::WaitForIncomingConnection(): not waiting !\n"));
		return EINVAL; // you can only wait for a connection if 
		// you didn't specify a node to connect to (by default the system audio input)
	}

	_m_lock.Lock();

	sem = _m_gotConnection;
	if (!m_isConnected && (sem <= B_OK))
		_m_gotConnection = sem = create_sem(0, "wait_for_audio_input");
	
	_m_lock.Unlock();
	
	if (sem > B_OK) {
		PRINT((BSNDREC"::WaitForIncomingConnection(): acquire_sem()\n"));
		err = acquire_sem_etc(sem, 1, block?0:B_RELATIVE_TIMEOUT, 0);
		if (err == B_TIMED_OUT)
			err = B_WOULD_BLOCK;
	} else
		return EINVAL;
	if (err < B_OK)
		return err;


	_m_lock.Lock();

	_m_gotConnection = -1;
	delete_sem(sem);
	if (format)
		*format = m_input.format.u.raw_audio;
	SetInitError(B_OK);
	_m_lock.Unlock();
	return B_OK;
}

void BSoundRecorder::SetInitError(status_t in_error)
{
	_m_init_err = in_error;
}


	
status_t BSoundRecorder::Start()
{
	status_t err = B_OK;
	bigtime_t then;
	
	PRINT((BSNDREC"Start()\n"));
	_m_lock.Lock();
	
	if (m_waitingForConnection && !m_isConnected) {
		err = EPIPE;
		goto bad1;
	}

	if (!m_isConnected) {
		//	Hook up input
		err = MakeRecordConnection(m_connectedTo);
		if (err < B_OK) {
			PRINT((BSNDREC"::Start(): connect to the selected sound input error %d\n", (int)err));
			goto bad1;
		}
		m_isConnected = true;
	}

	//	And get it going...
	then = _m_node->TimeSource()->Now()+50000LL;
	m_roster->StartNode(_m_node->Node(), then);
	if (m_connectedTo.kind & B_TIME_SOURCE) {
		m_roster->StartNode(m_connectedTo, _m_node->TimeSource()->RealTimeFor(then, 0));
	}
	else {
		m_roster->StartNode(m_connectedTo, then);
	}

	err = B_OK;
	
bad1:
	_m_lock.Unlock();

	return err;
}
				   
void BSoundRecorder::Stop(
						bool block,
						bool flush)
{
	PRINT((BSNDREC"Stop(%s, %s)\n", block?"true":"false", flush?"true":"false"));
	_m_lock.Lock();

	if (!m_isRecording)
		goto bad1;

	if (!m_waitingForConnection && m_isConnected) {
		BreakRecordConnection();
	}
bad1:
	_m_lock.Unlock();

	return;
}

void
BSoundRecorder::RecordBuffer(
	bigtime_t timestamp, 
	void * data,
	size_t size, 
	const media_multi_audio_format & format)
{
	PRINT((BSNDREC": RecordFile callback\n"));
	//	Callback called from the SoundConsumer when receiving buffers.

	if (_RecordFunc)
		(*_RecordFunc)(_m_cookie, timestamp, data, size, format);
}

void
BSoundRecorder::Notify(
//	void * cookie,
	sound_recorder_notification code,
	...)
{
	int32 cookie;
//	BSoundRecorder *inst = (BSoundRecorder *) cookie;
	PRINT((BSNDREC"Notify(%ld, ...)\n", (int32)code));
	switch (code) {
	case B_CONNECTED:
		_m_lock.Lock();
		cookie = 0;
		_m_node->GetNextInput(&cookie, &m_input);
		m_isConnected = true;
		if (_m_gotConnection > B_OK) {
			release_sem(_m_gotConnection);
		}
		_m_lock.Unlock();
		break;
	case B_DISCONNECTED:
		_m_lock.Lock();
		m_isConnected = false;
		_m_lock.Unlock();
		break;
	case B_FORMAT_CHANGED:
		_m_lock.Lock();
		cookie = 0;
		_m_node->GetNextInput(&cookie, &m_input);
		_m_lock.Unlock();
		break;
	case B_NODE_DIES:
//		Stop();
		_m_lock.Lock();
		m_isConnected = false;
		
		_m_lock.Unlock();
	default:
		return;
	}
	if (_NotifyFunc)
		(*_NotifyFunc)(_m_cookie, code);
}


status_t
BSoundRecorder::MakeRecordConnection(
	const media_node & input)
{
	PRINT((BSNDREC":MakeRecordConnection()\n"));
	
	//	Find an available output for the given input node.
	int32 count = 0;
	status_t err = m_roster->GetFreeOutputsFor(input, &m_output, 1, &count, B_MEDIA_RAW_AUDIO);
	if (err < B_OK) {
		PRINT((BSNDREC":MakeRecordConnection(): couldn't get free outputs from audio input node\n"));
		return err;
	}
	if (count < 1) {
		PRINT((BSNDREC":MakeRecordConnection(): no free outputs from audio input node\n"));
		return B_BUSY;
	}

	//	Find an available input for our own Node. Note that we go through the
	//	MediaRoster; calling Media Kit methods directly on Nodes in our app is
	//	not OK (because synchronization happens in the service thread, not in
	//	the calling thread).
	// TODO: explain this
	err = m_roster->GetFreeInputsFor(_m_node->Node(), &m_input, 1, &count, B_MEDIA_RAW_AUDIO);
	if (err < B_OK) {
		PRINT((BSNDREC":MakeRecordConnection(): couldn't get free inputs for sound recorder\n"));
		return err;
	}
	if (count < 1) {
		PRINT((BSNDREC":MakeRecordConnection(): no free inputs for sound recorder\n"));
		return B_BUSY;
	}

	//	Find out what the time source of the input is.
	//	For most nodes, we just use the preferred time source (the DAC) for synchronization.
	//	However, nodes that record from an input need to synchronize to the audio input node
	//	instead for best results.
	//	MakeTimeSourceFor gives us a "clone" of the time source node that we can manipulate
	//	to our heart's content. When we're done with it, though, we need to call Release()
	//	on the time source node, so that it keeps an accurate reference count and can delete
	//	itself when it's no longer needed.
	// TODO: what about filters connected to audio input?
	media_node use_time_source;
	BTimeSource * tsobj = m_roster->MakeTimeSourceFor(input);
	if (! tsobj) {
		PRINT((BSNDREC":MakeRecordConnection(): couldn't clone time source from audio input node\n"));
		return B_MEDIA_BAD_NODE;
	}

	//	Apply the time source in effect to our own Node.
	err = m_roster->SetTimeSourceFor(_m_node->Node().node, tsobj->Node().node);
	if (err < B_OK) {
		PRINT((BSNDREC":MakeRecordConnection(): couldn't set the sound recorder's time source\n"));
		tsobj->Release();
		return err;
	}

	//	Get a format, any format.
	media_format fmt;
	fmt.u.raw_audio = m_output.format.u.raw_audio;
	fmt.type = B_MEDIA_RAW_AUDIO;
	//test
/*
	fmt.u.raw_audio.frame_rate = 44100;
	fmt.u.raw_audio.channel_count = 2; 
	fmt.u.raw_audio.format = media_raw_audio_format::B_AUDIO_SHORT; 
*/

	//	Using the same structs for input and output is OK in BMediaRoster::Connect().
	err = m_roster->Connect(m_output.source, m_input.destination, &fmt, &m_output, &m_input);
	if (err < B_OK) {
		PRINT((BSNDREC":MakeRecordConnection(): failed to connect sound recorder to audio input node.\n"));
		tsobj->Release();
		return err;
	}

	//	Start the time source if it's not running.
	if ((tsobj->Node() != input) && !tsobj->IsRunning()) {
		m_roster->StartNode(tsobj->Node(), BTimeSource::RealTime());
	}
	tsobj->Release();	//	we're done with this time source instance!
	return B_OK;
}

status_t
BSoundRecorder::BreakRecordConnection()
{
	status_t err;

	_m_lock.Lock();

	if (!m_waitingForConnection && m_isConnected) {
		//	If we are the last connection, the Node will stop automatically since it
		//	has nowhere to send data to.
		err = m_roster->StopNode(m_input.node, 0);
		err = m_roster->Disconnect(m_output.node.node, m_output.source, m_input.node.node, m_input.destination);
		m_output.source = media_source::null;
		m_input.destination = media_destination::null;
	}

	_m_lock.Unlock();

	return err;
}

BSoundRecorder::SoundProcessFunc BSoundRecorder::BufferRecorder() const
{
	return _RecordFunc;
}

void BSoundRecorder::SetBufferRecorder(SoundProcessFunc recordFunc)
{
	_RecordFunc = recordFunc;
}

BSoundRecorder::SoundNotifyFunc BSoundRecorder::EventNotifier() const
{
	return _NotifyFunc;
}

void BSoundRecorder::SetNotifier(SoundNotifyFunc notifyFunc)
{
	_NotifyFunc = notifyFunc;
}

void *BSoundRecorder::Cookie() const
{
	return _m_cookie;
}

void BSoundRecorder::SetCookie(void *cookie)
{
	_m_cookie = cookie;
}

void BSoundRecorder::SetCallbacks(		//	atomic change
			SoundProcessFunc recordFunc,
			SoundNotifyFunc notifyFunc,
			void * cookie)
{
	_m_lock.Lock();
	_RecordFunc = recordFunc;
	_NotifyFunc = notifyFunc;
	_m_cookie = cookie;
	_m_lock.Unlock();	
}

bigtime_t BSoundRecorder::CurrentTime()
{
	return 0LL;
}

bigtime_t BSoundRecorder::PerformanceTime()
{
	return 0LL;
}

status_t BSoundRecorder::Preroll()
{
	return B_OK;
}

bool BSoundRecorder::HasData()
{
	return m_isConnected;
}

void BSoundRecorder::SetHasData(//	this does more than just set a flag
			bool has_data)
{
}


float BSoundRecorder::Volume()
{
	return 0.8;
}

void BSoundRecorder::SetVolume(float new_volume)
{

}

float BSoundRecorder::VolumeDB(bool forcePoll)
{
	if (!forcePoll && (system_time() < _m_gotVolume + 1000000))
		return 0.0;
	_m_gotVolume = system_time();
	return 0.0;
}

void BSoundRecorder::SetVolumeDB(float volume_dB)
{
	
}

status_t BSoundRecorder::GetVolumeInfo(
				media_node * out_node,
				int32 * out_parameter,
				float * out_min_dB,
				float * out_max_dB)
{
	return B_ERROR;
}

bigtime_t BSoundRecorder::Latency()
{
	return 0LL; /* UNIMPLEMENTED */
}

/* FBC provision */

status_t BSoundRecorder::_Reserved_SoundRecorder_0(void *, ...) { return B_ERROR; }
status_t BSoundRecorder::_Reserved_SoundRecorder_1(void *, ...) { return B_ERROR; }
status_t BSoundRecorder::_Reserved_SoundRecorder_2(void *, ...) { return B_ERROR; }
status_t BSoundRecorder::_Reserved_SoundRecorder_3(void *, ...) { return B_ERROR; }
status_t BSoundRecorder::_Reserved_SoundRecorder_4(void *, ...) { return B_ERROR; }
status_t BSoundRecorder::_Reserved_SoundRecorder_5(void *, ...) { return B_ERROR; }
status_t BSoundRecorder::_Reserved_SoundRecorder_6(void *, ...) { return B_ERROR; }
status_t BSoundRecorder::_Reserved_SoundRecorder_7(void *, ...) { return B_ERROR; }

} // Experimental
} // Media
} // BPrivate
