#include <stdio.h>
#include <stdlib.h>
#include <OS.h>
#include <Application.h>
#include <Messenger.h>
#include <SoundRecorder.h>
#include <SoundPlayer.h>

using namespace BPrivate::Media::Experimental;

void *audiobuffer = NULL;

sem_id buffsem = -1;

void MySoundPlayfunc(void *, void * buffer, size_t size, const media_raw_audio_format & format)
{
	acquire_sem(buffsem);
	memcpy(buffer, audiobuffer, size);
}

void MySoundProcessFunc(void * cookie,
	bigtime_t timestamp, void * data, size_t datasize,
	const media_multi_audio_format & format)
{
	printf("buffer\n");
	memcpy(audiobuffer, data, datasize);
	release_sem(buffsem);
}

int32 be_app_thread(void *)
{
	new BApplication("application/x-vnd.mmu_man.BSoundRecorder.test");
	be_app->Run();
	return B_OK;
}

int main(int argc, char **argv)
{
	status_t err;
	thread_id be_app_tid;
	BSoundRecorder *sr;
	BSoundPlayer *sp;
	int arg;
	bool wait_for_cnx = false;
	
	for (arg=1; arg < argc; arg++) {
		if (!strcmp(argv[arg], "-w"))
			wait_for_cnx = true;
		else if (!strcmp(argv[arg], "-h")) {
			puts("audio BSoundRecorder test: forwarding test.");
			printf("%s [-w]", argv[0]);
			puts("-w: wait for an input connection");
			return 0;
		}
	}
	
	resume_thread(be_app_tid = spawn_thread(be_app_thread, "BSoundRecorder.test.BApp", B_NORMAL_PRIORITY, NULL));
	while (!be_app)
		snooze(10000);

	media_format fmt;
	fmt.type = B_MEDIA_RAW_AUDIO;

	sr = new BSoundRecorder(&fmt.u.raw_audio, wait_for_cnx);
	sr->SetBufferRecorder(MySoundProcessFunc);
	printf("InitCheck() = 0x%08lx\n", sr->InitCheck());

	if (wait_for_cnx) {
		sr->WaitForIncomingConnection(&fmt.u.raw_audio);
	}


	char buff[100];
	fmt.u.raw_audio = sr->Format();
	string_for_format(fmt, buff, 98);
	printf("FMT= %s\n", buff);
	
	err = sr->Start();

	if (err < B_OK) {
		printf("error 0x%08lx\n", err);
		delete sr;
		be_app_messenger.SendMessage(new BMessage(B_QUIT_REQUESTED));
		wait_for_thread(be_app_tid, &err);
		free(audiobuffer);
		return 1;
	}
	audiobuffer = malloc(fmt.u.raw_audio.buffer_size);
	buffsem = create_sem(0, "audio_buffer_sem");

	sp = new BSoundPlayer(&fmt.u.raw_audio, "playback!", MySoundPlayfunc);

	sp->Start();
	sp->SetHasData(true);

	
	getchar();
	puts("***stopping***");
	delete_sem(buffsem);
	sp->Stop();
	sr->Stop();
	delete sp;
	delete sr;
	be_app_messenger.SendMessage(new BMessage(B_QUIT_REQUESTED));
	wait_for_thread(be_app_tid, &err);
	free(audiobuffer);
	return 0;
}
